from __future__ import annotations

from datetime import datetime
from datetime import timedelta

from optuna.storages import BaseStorage
from optuna.storages import RDBStorage
from optuna.study import StudyDirection
from optuna.study._frozen import FrozenStudy
from optuna.trial import FrozenTrial

from ._inmemory_cache import InMemoryCache


def get_trials(
    in_memory_cache: InMemoryCache, storage: BaseStorage, study_id: int
) -> list[FrozenTrial]:
    with in_memory_cache._trials_cache_lock:
        trials = in_memory_cache._trials_cache.get(study_id, None)

        # Not a big fan of the heuristic, but I can't think of anything better.
        if trials is None or len(trials) < 100:
            ttl_seconds = 2
        elif len(trials) < 500:
            ttl_seconds = 5
        else:
            ttl_seconds = 10

        last_fetched_at = in_memory_cache._trials_last_fetched_at.get(study_id, None)
        if (
            trials is not None
            and last_fetched_at is not None
            and datetime.now() - last_fetched_at < timedelta(seconds=ttl_seconds)
        ):
            return trials
    trials = storage.get_all_trials(study_id, deepcopy=False)

    with in_memory_cache._trials_cache_lock:
        in_memory_cache._trials_last_fetched_at[study_id] = datetime.now()
        in_memory_cache._trials_cache[study_id] = trials
    return trials


def get_studies(storage: BaseStorage) -> list[FrozenStudy]:
    frozen_studies = storage.get_all_studies()
    if isinstance(storage, RDBStorage):
        frozen_studies = sorted(frozen_studies, key=lambda s: s._study_id)
    return frozen_studies


def get_study(storage: BaseStorage, study_id: int) -> FrozenStudy | None:
    studies = get_studies(storage)
    for s in studies:
        if s._study_id != study_id:
            continue
        return s
    return None


def create_new_study(
    storage: BaseStorage, study_name: str, directions: list[StudyDirection]
) -> int:
    study_id = storage.create_new_study(directions, study_name=study_name)
    return study_id



from __future__ import annotations

from typing import Any
import uuid

from optuna.storages import BaseStorage
from optuna.trial import TrialState


_SYSTEM_ATTR_PREFIX_PREFERENCE = "preference:values"
_SYSTEM_ATTR_PREFIX_SKIP_TRIAL = "preference:skip_trial:"
_SYSTEM_ATTR_N_GENERATE = "preference:n_generate"


def report_preferences(
    study_id: int,
    storage: BaseStorage,
    preferences: list[tuple[int, int]],
) -> str:
    preference_id = str(uuid.uuid4())
    key = _SYSTEM_ATTR_PREFIX_PREFERENCE + preference_id
    storage.set_study_system_attr(
        study_id=study_id,
        key=key,
        value=preferences,
    )
    trials = storage.get_all_trials(study_id, deepcopy=False)
    directions = storage.get_study_directions(study_id)
    values = [0 for _ in directions]
    updated_trials = {num for tpl in preferences for num in tpl}
    for number in updated_trials:
        trial_id = trials[number]._trial_id
        if trials[number].state != TrialState.COMPLETE:
            storage.set_trial_state_values(trial_id, TrialState.COMPLETE, values)
    return preference_id


def get_preferences(study_system_attrs: dict[str, Any]) -> list[tuple[int, int]]:
    preferences: list[tuple[int, int]] = []
    for k, v in study_system_attrs.items():
        if not k.startswith(_SYSTEM_ATTR_PREFIX_PREFERENCE):
            continue
        preferences.extend(v)  # type: ignore
    return preferences


def is_preference_removed(study_system_attrs: dict[str, Any], preference_id: str) -> bool:
    key = _SYSTEM_ATTR_PREFIX_PREFERENCE + preference_id
    preference = study_system_attrs.get(key, [])
    return len(preference) == 0


def report_skip(
    study_id: int,
    trial_id: int,
    storage: BaseStorage,
) -> None:
    storage.set_study_system_attr(
        study_id=study_id,
        key=_SYSTEM_ATTR_PREFIX_SKIP_TRIAL + str(trial_id),
        value=True,
    )


def is_skipped_trial(trial_id: int, study_system_attrs: dict[str, Any]) -> bool:
    key = _SYSTEM_ATTR_PREFIX_SKIP_TRIAL + str(trial_id)
    return key in study_system_attrs


def get_skipped_trial_ids(study_system_attrs: dict[str, Any]) -> list[int]:
    skipped_trial_ids: list[int] = []
    for k in study_system_attrs:
        if not k.startswith(_SYSTEM_ATTR_PREFIX_SKIP_TRIAL):
            continue
        try:
            trial_id = int(k[len(_SYSTEM_ATTR_PREFIX_SKIP_TRIAL) :])  # noqa: E203
            skipped_trial_ids.append(trial_id)
        except ValueError:
            continue
    return skipped_trial_ids


def get_n_generate(study_system_attrs: dict[str, Any]) -> int:
    return study_system_attrs[_SYSTEM_ATTR_N_GENERATE]


def set_n_generate(study_id: int, storage: BaseStorage, n_generate: int) -> None:
    storage.set_study_system_attr(
        study_id=study_id,
        key=_SYSTEM_ATTR_N_GENERATE,
        value=n_generate,
    )



from ._app import run_server  # noqa
from ._app import wsgi  # noqa
from ._custom_plot_data import save_plotly_graph_object  # noqa
from ._form_widget import ChoiceWidget  # noqa
from ._form_widget import dict_to_form_widget  # noqa
from ._form_widget import ObjectiveChoiceWidget  # noqa
from ._form_widget import ObjectiveSliderWidget  # noqa
from ._form_widget import ObjectiveTextInputWidget  # noqa
from ._form_widget import ObjectiveUserAttrRef  # noqa
from ._form_widget import register_objective_form_widgets  # noqa
from ._form_widget import register_user_attr_form_widgets  # noqa
from ._form_widget import SliderWidget  # noqa
from ._form_widget import TextInputWidget  # noqa
from ._named_objectives import set_objective_names  # noqa
from ._note import get_note  # noqa
from ._note import save_note  # noqa
from ._preference_setting import register_preference_feedback_component  # noqa


__version__ = "0.19.0"



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
import json
from typing import TYPE_CHECKING

from optuna.storages import BaseStorage

from .preferential._system_attrs import _SYSTEM_ATTR_PREFIX_PREFERENCE
from .preferential._system_attrs import report_preferences


_SYSTEM_ATTR_PREFIX_HISTORY = "preference:history"

if TYPE_CHECKING:
    from typing import Literal
    from typing import TypedDict

    FeedbackMode = Literal["ChooseWorst"]
    ChooseWorstHistory = TypedDict(
        "ChooseWorstHistory",
        {
            "mode": FeedbackMode,
            "id": str,
            "timestamp": str,
            "candidates": list[int],
            "clicked": int,
            "preferences": list[tuple[int, int]],
        },
    )
    History = ChooseWorstHistory
    SerializedHistory = TypedDict(
        "SerializedHistory",
        {
            "history": History,
            "is_removed": bool,
        },
    )


class PreferenceHistoryNotFound(Exception):
    pass


@dataclass
class NewHistory:
    mode: FeedbackMode
    candidates: list[int]
    clicked: int


def report_history(
    study_id: int,
    storage: BaseStorage,
    input_data: NewHistory,
) -> str:
    preferences = []
    # TODO(moririn): Use TypeGuard after adding other history types.
    if input_data.mode == "ChooseWorst":
        preferences = [
            (better, input_data.clicked)
            for better in input_data.candidates
            if better != input_data.clicked
        ]
    else:
        assert False, f"Unknown data: {input_data}"

    preference_id = report_preferences(
        study_id=study_id,
        storage=storage,
        preferences=preferences,
    )

    if input_data.mode == "ChooseWorst":
        history: ChooseWorstHistory = {
            "mode": "ChooseWorst",
            "id": preference_id,
            "timestamp": datetime.now().isoformat(),
            "candidates": input_data.candidates,
            "clicked": input_data.clicked,
            "preferences": preferences,
        }

    key = _SYSTEM_ATTR_PREFIX_HISTORY + preference_id
    storage.set_study_system_attr(
        study_id=study_id,
        key=key,
        value=json.dumps(history),
    )
    return preference_id


def remove_history(study_id: int, storage: BaseStorage, history_id: str) -> None:
    system_attrs = storage.get_study_system_attrs(study_id)
    history_key = _SYSTEM_ATTR_PREFIX_HISTORY + history_id
    if history_key not in system_attrs:
        raise PreferenceHistoryNotFound
    storage.set_study_system_attr(study_id, _SYSTEM_ATTR_PREFIX_PREFERENCE + history_id, [])


def restore_history(study_id: int, storage: BaseStorage, history_id: str) -> None:
    system_attrs = storage.get_study_system_attrs(study_id)
    history_key = _SYSTEM_ATTR_PREFIX_HISTORY + history_id
    if history_key not in system_attrs:
        raise PreferenceHistoryNotFound
    history: History = json.loads(system_attrs.get(history_key, ""))
    storage.set_study_system_attr(
        study_id, _SYSTEM_ATTR_PREFIX_PREFERENCE + history_id, history["preferences"]
    )



from __future__ import annotations

import os.path
from pathlib import Path
import re
from typing import TYPE_CHECKING

from optuna.storages import BaseStorage
from optuna.storages import RDBStorage
from optuna.version import __version__ as optuna_ver
from packaging import version


if TYPE_CHECKING:
    from typing import Optional
    from typing import Union

    from optuna.storages import JournalStorage


# The code to declare `rfc1738_pattern` variable is quoted from the source code
# of SQLAlchemy project (MIT License). See LICENSE file for detials.
# https://github.com/zzzeek/sqlalchemy/blob/c6554ac52/lib/sqlalchemy/engine/url.py#L234-L292
rfc1738_pattern = re.compile(
    r"""
        (?P<name>[\w\+]+)://
        (?:
            (?P<username>[^:/]*)
            (?::(?P<password>.*))?
        @)?
        (?:
            (?:
                \[(?P<ipv6host>[^/]+)\] |
                (?P<ipv4host>[^/:]+)
            )?
            (?::(?P<port>[^/]*))?
        )?
        (?:/(?P<database>.*))?
        """,
    re.X,
)


def get_storage(
    storage: Union[str, BaseStorage], storage_class: Optional[str] = None
) -> BaseStorage:
    if isinstance(storage, BaseStorage):
        return storage

    if storage_class:
        if storage_class == "RDBStorage":
            return get_rdb_storage(storage)
        if storage_class == "JournalRedisStorage":
            return get_journal_redis_storage(storage)
        if storage_class == "JournalFileStorage":
            return get_journal_file_storage(storage)
        raise ValueError("Unexpected storage_class")

    return guess_storage_from_url(storage)


def _has_sqlite_header(storage_url: str) -> bool:
    storage_path = Path(storage_url)
    SQLITE_HEADER = (
        b"SQLite format 3\x00"  # see https://github.com/optuna/optuna-dashboard/pull/800
    )
    with storage_path.open(mode="rb") as f:
        header = f.read(len(SQLITE_HEADER))
    return header == SQLITE_HEADER


def guess_storage_from_url(storage_url: str) -> BaseStorage:
    if storage_url.startswith("redis"):
        return get_journal_redis_storage(storage_url)

    if os.path.isfile(storage_url):
        if _has_sqlite_header(storage_url):
            raise ValueError(
                f"Please specify 'sqlite:///{storage_url}' to use SQLite3 (RDBStorage)"
            )
        return get_journal_file_storage(storage_url)

    if rfc1738_pattern.match(storage_url) is not None:
        return get_rdb_storage(storage_url)

    raise ValueError("Failed to guess storage class from storage_url")


def get_rdb_storage(storage_url: str) -> RDBStorage:
    if version.parse(optuna_ver) >= version.Version("v3.0.0"):
        return RDBStorage(storage_url, skip_compatibility_check=True, skip_table_creation=True)
    else:
        return RDBStorage(storage_url, skip_compatibility_check=True)


def get_journal_file_storage(file_path: str) -> JournalStorage:
    if version.parse(optuna_ver) < version.Version("v3.1.0"):
        raise ValueError("JournalRedisStorage is available from Optuna v3.1.0")

    from optuna.storages import JournalFileOpenLock
    from optuna.storages import JournalStorage

    try:
        from optuna.storages.journal import JournalFileBackend
    except ImportError:
        from optuna.storages import JournalFileStorage as JournalFileBackend

    storage: JournalStorage
    if os.name == "nt":
        lock_obj = JournalFileOpenLock(file_path)
        storage = JournalStorage(JournalFileBackend(file_path=file_path, lock_obj=lock_obj))
    else:
        storage = JournalStorage(JournalFileBackend(file_path=file_path))
    return storage


def get_journal_redis_storage(redis_url: str) -> JournalStorage:
    if version.parse(optuna_ver) < version.Version("v3.1.0"):
        raise ValueError("JournalRedisStorage is available from Optuna v3.1.0")

    from optuna.storages import JournalRedisStorage
    from optuna.storages import JournalStorage

    return JournalStorage(JournalRedisStorage(redis_url))



from __future__ import annotations

from datetime import datetime
import numbers
import threading
from typing import cast
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

from optuna.distributions import BaseDistribution
from optuna.distributions import CategoricalDistribution
from optuna.distributions import FloatDistribution
from optuna.distributions import IntDistribution
from optuna.trial import FrozenTrial
from optuna.trial import TrialState


if TYPE_CHECKING:
    SearchSpaceSetT = Set[Tuple[str, BaseDistribution]]
    SearchSpaceListT = List[Tuple[str, BaseDistribution]]


def get_cached_extra_study_property(
    in_memory_cache: InMemoryCache, study_id: int, trials: list[FrozenTrial]
) -> tuple[SearchSpaceListT, SearchSpaceListT, list[tuple[str, bool]], bool]:
    with in_memory_cache._cached_extra_study_property_cache_lock:
        cached_extra_study_property = in_memory_cache._cached_extra_study_property_cache.get(
            study_id, None
        )
        if cached_extra_study_property is None:
            cached_extra_study_property = _CachedExtraStudyProperty()
        cached_extra_study_property.update(trials)
        in_memory_cache._cached_extra_study_property_cache[study_id] = cached_extra_study_property
        return (
            cached_extra_study_property.intersection_search_space,
            cached_extra_study_property.union_search_space,
            cached_extra_study_property.union_user_attrs,
            cached_extra_study_property.has_intermediate_values,
        )


class InMemoryCache:
    def __init__(self) -> None:
        self._cached_extra_study_property_cache: dict[int, "_CachedExtraStudyProperty"] = {}
        self._cached_extra_study_property_cache_lock = threading.Lock()
        self._trials_cache: dict[int, list[FrozenTrial]] = {}
        self._trials_cache_lock = threading.Lock()
        self._trials_last_fetched_at: dict[int, datetime] = {}

    def clear(self) -> None:
        with self._cached_extra_study_property_cache_lock:
            self._cached_extra_study_property_cache.clear()
        with self._trials_cache_lock:
            self._trials_cache.clear()
            self._trials_last_fetched_at.clear()


class _CachedExtraStudyProperty:
    def __init__(self) -> None:
        self._cursor: int = -1
        self._intersection_search_space: Optional[SearchSpaceSetT] = None
        self._union_search_space: SearchSpaceSetT = set()
        self._union_user_attrs: dict[str, bool] = {}  # attr_name: is_sortable (= is_number)
        self.has_intermediate_values: bool = False

    @property
    def intersection_search_space(self) -> SearchSpaceListT:
        if self._intersection_search_space is None:
            return []
        intersection = list(self._intersection_search_space)
        intersection.sort(key=lambda x: x[0])
        return intersection

    @property
    def union_search_space(self) -> SearchSpaceListT:
        union = list(self._union_search_space)
        union.sort(key=lambda x: x[0])
        return union

    @property
    def union_user_attrs(self) -> list[tuple[str, bool]]:
        union = [(name, is_sortable) for name, is_sortable in self._union_user_attrs.items()]
        sorted(union, key=lambda x: x[0])
        return union

    def update(self, trials: list[FrozenTrial]) -> None:
        next_cursor = self._cursor
        for trial in reversed(trials):
            if self._cursor > trial.number:
                break

            if not trial.state.is_finished():
                next_cursor = trial.number

            self._update_user_attrs(trial)
            if trial.state != TrialState.FAIL:
                self._update_intermediate_values(trial)
                self._update_search_space(trial)

        self._cursor = next_cursor

    def _update_user_attrs(self, trial: FrozenTrial) -> None:
        current_user_attrs = {
            k: not isinstance(v, bool) and isinstance(v, numbers.Real)
            for k, v in trial.user_attrs.items()
        }
        for attr_name, current_is_sortable in current_user_attrs.items():
            is_sortable = self._union_user_attrs.get(attr_name)
            if is_sortable is None:
                self._union_user_attrs[attr_name] = current_is_sortable
            elif is_sortable and not current_is_sortable:
                self._union_user_attrs[attr_name] = False

    def _update_intermediate_values(self, trial: FrozenTrial) -> None:
        if not self.has_intermediate_values and len(trial.intermediate_values) > 0:
            self.has_intermediate_values = True

    def _update_search_space(self, trial: FrozenTrial) -> None:
        def _is_same_float_distribution(d1: BaseDistribution, d2: BaseDistribution) -> bool:
            return (
                isinstance(d1, FloatDistribution)
                and isinstance(d2, FloatDistribution)
                and d1.step == d2.step
                and d1.log == d2.log
            )

        def _is_same_int_distribution(d1: BaseDistribution, d2: BaseDistribution) -> bool:
            return (
                isinstance(d1, IntDistribution)
                and isinstance(d2, IntDistribution)
                and d1.step == d2.step
                and d1.log == d2.log
            )

        def _is_same_categorical_distribution(d1: BaseDistribution, d2: BaseDistribution) -> bool:
            return isinstance(d1, CategoricalDistribution) and isinstance(
                d2, CategoricalDistribution
            )

        new_union_search_space: SearchSpaceSetT = set(
            [(n, d) for n, d in self._union_search_space]
        )
        for n1, d1 in trial.distributions.items():
            for n2, d2 in self._union_search_space:
                if n1 == n2 and _is_same_float_distribution(d1, d2):
                    d1, d2 = cast(FloatDistribution, d1), cast(FloatDistribution, d2)
                    new_float_search_space = (
                        n1,
                        FloatDistribution(
                            low=min(d1.low, d2.low),
                            high=max(d1.high, d2.high),
                            step=d1.step,
                            log=d1.log,
                        ),
                    )
                    new_union_search_space.remove((n2, d2))
                    new_union_search_space.add(new_float_search_space)
                    break
                elif n1 == n2 and _is_same_int_distribution(d1, d2):
                    d1, d2 = cast(IntDistribution, d1), cast(IntDistribution, d2)
                    new_int_search_space = (
                        n1,
                        IntDistribution(
                            low=min(d1.low, d2.low),
                            high=max(d1.high, d2.high),
                            step=d1.step,
                            log=d1.log,
                        ),
                    )
                    new_union_search_space.remove((n2, d2))
                    new_union_search_space.add(new_int_search_space)
                    break
                elif n1 == n2 and _is_same_categorical_distribution(d1, d2):
                    d1, d2 = cast(CategoricalDistribution, d1), cast(CategoricalDistribution, d2)
                    new_categorical_search_space = (
                        n1,
                        CategoricalDistribution(choices=list(set(d1.choices + d2.choices))),
                    )
                    new_union_search_space.remove((n2, d2))
                    new_union_search_space.add(new_categorical_search_space)
                    break
            else:
                new_union_search_space.add((n1, d1))
        self._union_search_space = new_union_search_space

        if self._intersection_search_space is None:
            self._intersection_search_space = set([(n, d) for n, d in trial.distributions.items()])
        else:
            new_intersection_search_space: SearchSpaceSetT = set()
            for n1, d1 in trial.distributions.items():
                for n2, d2 in self._intersection_search_space:
                    if n1 == n2 and _is_same_float_distribution(d1, d2):
                        d1, d2 = cast(FloatDistribution, d1), cast(FloatDistribution, d2)
                        new_float_search_space = (
                            n1,
                            FloatDistribution(
                                low=min(d1.low, d2.low),
                                high=max(d1.high, d2.high),
                                step=d1.step,
                                log=d1.log,
                            ),
                        )
                        new_intersection_search_space.add(new_float_search_space)
                        break
                    elif n1 == n2 and _is_same_int_distribution(d1, d2):
                        d1, d2 = cast(IntDistribution, d1), cast(IntDistribution, d2)
                        new_int_search_space = (
                            n1,
                            IntDistribution(
                                low=min(d1.low, d2.low),
                                high=max(d1.high, d2.high),
                                step=d1.step,
                                log=d1.log,
                            ),
                        )
                        new_intersection_search_space.add(new_int_search_space)
                        break
                    elif n1 == n2 and _is_same_categorical_distribution(d1, d2):
                        d1, d2 = (
                            cast(CategoricalDistribution, d1),
                            cast(CategoricalDistribution, d2),
                        )
                        new_categorical_search_space = (
                            n1,
                            CategoricalDistribution(choices=list(set(d1.choices + d2.choices))),
                        )
                        new_intersection_search_space.add(new_categorical_search_space)
                        break
            self._intersection_search_space = new_intersection_search_space
